{- This module -}
   
module HarmTrace.Harmony.ChordRep where

-- HarmTrace
import           HarmTrace.Base.MusicRep
import           HarmTrace.Models.Jazz.Main
import           HarmTrace.Audio.BeatChroma
-- Haskore
import qualified Haskore.Basic.Pitch as Pitch (Class(..), Relative, fromInt, Octave)
import           Haskore.Basic.Scale 
-- Misc
import Data.List

type NoteInt = Int     
type Index   = Int

allShortHands :: [Shorthand]
allShortHands = [sh | sh <- [Maj .. None]]

allAdditions :: [Addition]
allAdditions = [Note (Just modifier) interval | modifier <- [Sh .. FF], interval <- [I1 .. I13]]

allChordScaleDegree :: [Chord ScaleDegree]
allChordScaleDegree = [(Chord scaledegree chordShorthand [] getLoc duration) |  scaledegree <- scaleDegrees
                                                                               , chordShorthand <- [Maj .. Sus2]
                                                                               --, chordAdditions <- allAdditions
                                                                               , getLoc <- [0]
                                                                               , duration <- [0] ]

                            --chord                chord, double representation
allChordScaleDegreeStruc :: [Chord ScaleDegree] -> [(Chord ScaleDegree,[Double])]
allChordScaleDegreeStruc (x:xs) = (x,(chordScaleDegreeToChordStructure x)):allChordScaleDegreeStruc xs
allChordScaleDegreeStruc []     = []

-- returns all chords which have 1 on position num in their double representation
whichChordsHaveInt :: Int -> Index -> [(Chord ScaleDegree,[Double])] -> [Chord ScaleDegree]
whichChordsHaveInt num idx (((Chord (Note (a) b) sh [] 0 0 ),struc):shs) | ((struc!!num)==1.0) = (Chord (Note (a) b) sh [] idx 0):(whichChordsHaveInt num idx shs)
                                                                         | otherwise           = whichChordsHaveInt num idx shs 
whichChordsHaveInt num _ []       = []
                                                                  
findChordsByInt :: (Int,Double) -> Key -> Index -> [Chord ScaleDegree]
findChordsByInt (i,d) (Key (Note (_) n ) mode) idx = 
        case mode of 
             MajMode -> filter isChordInMaj $ whichChordsHaveInt i idx (allChordScaleDegreeStruc allChordScaleDegree)  
             MinMode -> filter isChordInMin $ whichChordsHaveInt i idx (allChordScaleDegreeStruc allChordScaleDegree)

findChordsByInt2 :: Int -> Key -> Index -> [Chord ScaleDegree]
findChordsByInt2 (-1) (Key (Note (_) n ) mode) idx = []
findChordsByInt2 i (Key (Note (_) n ) mode) idx = 
        case mode of 
             MajMode -> filter isChordInMaj $ whichChordsHaveInt i idx (allChordScaleDegreeStruc allChordScaleDegree)  
             MinMode -> filter isChordInMin $ whichChordsHaveInt i idx (allChordScaleDegreeStruc allChordScaleDegree)

--                key    melody              chords                     chords with probability
songChordProps :: Key -> [[(Int,Double)]] -> [[[Chord ScaleDegree]]] -> [[[(Chord ScaleDegree,Double)]]]
songChordProps key (x:xs) (y:ys) = barChordProps key x y: songChordProps key xs ys
songChordProps key [] _          = []

barChordProps :: Key -> [(Int,Double)] -> [[Chord ScaleDegree]] -> [[(Chord ScaleDegree,Double)]]
barChordProps key (m:ms) (c:cs) = noteChordProps key m c:barChordProps key ms cs
barChordProps key [] _          = []

noteChordProps :: Key -> (Int,Double) -> [Chord ScaleDegree] -> [(Chord ScaleDegree,Double)]
noteChordProps key (n,d) (c:cs) = (c,(getProp key c n)):noteChordProps key (n,d) (cs)
noteChordProps key (n,d) []     = []

-- takes a key, chord and note as int, and returns 1 - (distance in the circle of fifths /10 )
getProp :: Key -> Chord ScaleDegree -> Int -> Double
getProp key@(Key (Note (_) sd ) MajMode) c n = abs(1 - prop) where 
          prop = ((noteMapdiff (rootFromScaleDegree key c) ( (getScaleFromKey key)!!(convergeNotes n majRel 0 0) ))/10)
getProp key@(Key (Note (_) sd ) MinMode) c n = abs(1 - prop) where
          prop = ((noteMapdiff (rootFromScaleDegree key c) ( (getScaleFromKey key)!!(convergeNotes n minRel 0 0) ))/10)

-- convergeNotes converts from an absolute chromatic note number to a relative 
--                n     minrel   0      0
convergeNotes :: Int -> [Int] -> Int -> Int -> Int
convergeNotes n (x:xs) a pos | (a == n)  = pos
                             | otherwise = convergeNotes n xs (a+x) (pos+1)

-- relative intervals of maj and min keys
majRel, minRel :: [Int]                   
majRel     = [ 2, 2, 1, 2, 2, 2, 1 ]
minRel     = [ 2, 1, 2, 2, 1, 2, 2 ]   

getBestHarm :: [[[(Chord ScaleDegree,Double)]]] -> [Chord ScaleDegree]
getBestHarm chords = concat $ (map.map) getBestChord chords

getBestChord :: [(Chord ScaleDegree,Double)] -> Chord ScaleDegree
getBestChord (x:xs) = getBestChord' (x:xs) (Chord (Note (Nothing) I) Maj [] 0 0) 0 where
        getBestChord' :: [(Chord ScaleDegree,Double)] -> Chord ScaleDegree -> Double -> Chord ScaleDegree
        getBestChord' ((c,s):xs) k a | (s >= a) = getBestChord' xs c s
                                     | (s <= a) = getBestChord' xs k a
        getBestChord' [] k a         = k

beatFilter :: (Double,Double) -> [Chord ScaleDegree] -> Int -> [Chord ScaleDegree]
beatFilter beat@(q,r) chords@(x:xs) n | (barStructure(beat)!!(mod n (round(q)))) = x:beatFilter beat xs (n+1)
                                      | otherwise                                = (Chord (Note (Nothing) Imp) None [] 0 0):beatFilter beat xs (n+1)
beatFilter beat@(q,r) [] n            = []
  

rootFromScaleDegree :: Key -> Chord ScaleDegree -> Root
rootFromScaleDegree key sd = (Note (modifier) root) where
                             (Chord (Note (modifier) root) sh l g d) = chordScaleDegreeToRoot key sd

-- ALSO ADD THE ADDITIONS: SET APPROPRIATE VALUE ON 1
-- convert a Chord ScaleDegree to a list of doubles, representing the used relative notes
chordScaleDegreeToChordStructure :: Chord ScaleDegree -> [Double]
chordScaleDegreeToChordStructure (Chord (Note (modifier) degree) chordShorthand chordAdditions getLoc duration) = 
        rotate (degreeToInt degree) (shortHandToChordStructure chordShorthand)

-- represent a shorthand as a list of doubles, wherein a double represents the used relative note
shortHandToChordStructure :: Shorthand -> [Double]
shortHandToChordStructure sh = case sh of                  
  --root ex. C,  Db, D,  Eb, E,  F,  F#, G,  Ab, A,  Bb,  B
  --relat.   0,  1,  2,  3,  4,  5,  6,  7,  8,  9,  10,  11
  Maj     -> [1,  0,  0,  0,  1,  0,  0,  1,  0,  0,  0,   0 ]
  Min     -> [1,  0,  0,  1,  0,  0,  0,  1,  0,  0,  0,   0 ]
  HDim7   -> [1,  0,  0,  1,  0,  0,  1,  0,  0,  0,  1,   0 ]
  Dim     -> [1,  0,  0,  1,  0,  0,  1,  0,  0,  0,  0,   0 ]
  Maj6    -> [1,  0,  0,  0,  1,  0,  0,  1,  0,  1,  0,   0 ]
  Sev     -> [1,  0,  0,  0,  1,  0,  0,  1,  0,  0,  1,   0 ]
  Dim7    -> [1,  0,  0,  1,  0,  0,  1,  0,  0,  1,  0,   0 ]
  Maj7    -> [1,  0,  0,  0,  1,  0,  0,  1,  0,  0,  0,   1 ]
  Min7    -> [1,  0,  0,  1,  0,  0,  0,  1,  0,  0,  1,   0 ]
  Min6    -> [1,  0,  0,  1,  0,  0,  0,  1,  0,  1,  0,   0 ]
  Sus4    -> [1,  0,  0,  0,  0,  1,  0,  1,  0,  0,  0,   0 ]
  -- Problematic? 
  Aug     -> [1,  0,  0,  0,  1,  0,  0,  0,  1,  0,  0,   0 ]
  MinMaj7 -> [1,  0,  0,  1,  0,  0,  0,  1,  0,  0,  0,   1 ]
  Maj9    -> [1,  1,  0,  0,  1,  0,  0,  1,  0,  0,  0,   0 ]  -- Ninth should be an octave higher...
  Min9    -> [1,  1,  0,  1,  0,  0,  0,  1,  0,  0,  0,   0 ]
  Sus2    -> [1,  0,  1,  0,  0,  0,  0,  1,  0,  0,  0,   0 ]
--Nin     -> Impossible?
  None    -> [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0 ]
  o       -> [0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,   0 ]

noteMapdiff :: Note DiatonicNatural -> Note DiatonicNatural -> Double
noteMapdiff key1 key2 = abs(pos1 - pos2) where 
                             [pos1] = findAltLists key1
                             [pos2] = findAltLists key2
-- | (abs(pos1 - pos2)) > 6 = 12 - abs(pos1 - pos2)
-- | otherwise              = abs(pos1 - pos2) where

findAltLists :: Note DiatonicNatural -> [Double]
findAltLists key | (findPos circleOfFifths key == []) = findPos circleOfFifths' key
                 | otherwise                          = findPos circleOfFifths key
                    
-- rotate a list by n
rotate :: Int -> [a] -> [a]
rotate x st = take (length st) $ drop (negate x `mod` length st) $ cycle st

-- return the positions of an element in a list as a list
findPos :: (Num b, Eq b1, Enum b) => [b1] -> b1 -> [b]
findPos list elt = map fst $ filter ((elt==).snd) $ zip [0..] list

degreeToInt :: DiatonicDegree -> Int
degreeToInt degree = case degree of 
        I   -> 0
        II  -> 2
        III -> 4
        IV  -> 5
        V   -> 7
        VI  -> 9
        VII -> 11
        o   -> 12

-- Make a list of strong and weak bar parts with a time signature
barStructure :: (Double,Double) -> [Bool]
barStructure (0.0,d) = []
barStructure (1.0,d) = [True]
barStructure (n,d) = barStructure' (n,d) 0
barStructure' :: (Double,Double) -> Int -> [Bool]
barStructure' (n,d) i | ((i == 0) || ((i == floor (n/2.0)) && (i /= 1))) = True:barStructure' (n,d) (i+1)
                      | ((round(n)-1) == (i))                            = [False]
                      | otherwise                                        = False:barStructure' (n,d) (i+1)

circleOfFifths :: [Note DiatonicNatural]  
circleOfFifths =  [ Note Nothing   C 
                  , Note Nothing   G 
                  , Note Nothing   D
                  , Note Nothing   A
                  , Note Nothing   E
                  , Note Nothing   B
                  , Note (Just Sh) F 
                  , Note (Just Sh) C
                  , Note (Just Sh) G
                  , Note (Just Sh) D
                  , Note (Just Fl) B   
                  , Note Nothing   F ]            

circleOfFifths' :: [Note DiatonicNatural]  
circleOfFifths' =  [ Note (Just Sh) B 
                   , Note Nothing   G 
                   , Note Nothing   D
                   , Note Nothing   A
                   , Note Nothing   E
                   , Note Nothing   B
                   , Note (Just Fl) G 
                   , Note (Just Fl) D
                   , Note (Just Fl) A
                   , Note (Just Fl) E
                   , Note (Just Sh) A   
                   , Note Nothing   F ]  

-- check if a specific chord is a major key chord // M m m M M m dim         
isChordInMaj :: Chord ScaleDegree -> Bool
isChordInMaj chord = case chord of
        Chord (Note (Nothing) I)   Maj  _ _ _ -> True
        Chord (Note (Nothing) II)  Min  _ _ _ -> True
        Chord (Note (Nothing) III) Min  _ _ _ -> True
        Chord (Note (Nothing) IV)  Maj  _ _ _ -> True
        Chord (Note (Nothing) V)   Maj  _ _ _ -> True
        Chord (Note (Nothing) VI)  Min  _ _ _ -> True
        Chord (Note (Nothing) VII) Dim  _ _ _ -> True        
        o                                     -> False

-- check if a specific chord is a minor key chord // m dim M m m M M
isChordInMin :: Chord ScaleDegree -> Bool
isChordInMin chord = case chord of
        Chord (Note (Nothing) I)   Min  _ _ _ -> True
        Chord (Note (Nothing) II)  Dim  _ _ _ -> True
        Chord (Note (Nothing) III) Maj  _ _ _ -> True
        Chord (Note (Nothing) IV)  Min  _ _ _ -> True
        Chord (Note (Nothing) V)   Min  _ _ _ -> True
        Chord (Note (Nothing) VI)  Maj  _ _ _ -> True
        Chord (Note (Nothing) VII) Maj  _ _ _ -> True        
        o                                     -> False

--getScaleFromKey :: Key -> [(Haskore.Basic.Pitch.Octave, Class)]
getScaleFromKey :: Key -> [Root]
getScaleFromKey key = init $ map tupelToRoot (map Pitch.fromInt $ getIntervalsKey key)

-- chord modes of maj and min keys
majChordRel, minChordRel :: [Shorthand]
majChordRel = [Maj, Min, Min, Maj, Maj, Min, Dim]
minChordRel = [Min, Dim, Maj, Min, Min, Maj, Maj]

tupelToRoot :: (Pitch.Octave, Pitch.Class) -> Root
tupelToRoot (_,c) = classToRoot(c) 

-- convert a Chord ScaleDegree to a Chord Root by a Key
chordScaleDegreeToRoot :: Key -> Chord ScaleDegree -> Chord Root
chordScaleDegreeToRoot key@(Key (Note (_) _) _) (Chord (Note (_) Imp) None add getLoc duration)  = (Chord (Note (Nothing) N) None add getLoc duration)
chordScaleDegreeToRoot key@(Key (Note (modifier) n) MajMode) (Chord (Note (_) scaledegree) chordShorthand add getLoc duration)  = 
        (Chord ((getScaleFromKey(key))!!i) (majChordRel!!i) add getLoc duration)
        where i = degreeToAbs(scaledegree)
chordScaleDegreeToRoot key@(Key (Note (modifier) n) MinMode) (Chord (Note (_) scaledegree) chordShorthand add getLoc duration) = 
        (Chord ((getScaleFromKey(key))!!i) (minChordRel!!i) add getLoc duration)
        where i = degreeToAbs(scaledegree)

-- return the relative intervals of a key
getIntervalsKey :: Key -> [Int]
getIntervalsKey (Key root MajMode) = (fromIntervals majRel) (rootToClass(root))
getIntervalsKey (Key root MinMode) = (fromIntervals minRel) (rootToClass(root))

rootToClass :: Root -> Pitch.Class
rootToClass root = case root of
        Note (Just Fl) C -> Pitch.Cf 
        Note (Nothing) C -> Pitch.C 
        Note (Just Sh) C -> Pitch.Cs 
        Note (Just Fl) D -> Pitch.Df 
        Note (Nothing) D -> Pitch.D 
        Note (Just Sh) D -> Pitch.Ds  
        Note (Just Fl) E -> Pitch.Ef 
        Note (Nothing) E -> Pitch.E 
        Note (Just Sh) E -> Pitch.Es  
        Note (Just Fl) F -> Pitch.Ff 
        Note (Nothing) F -> Pitch.F 
        Note (Just Sh) F -> Pitch.Fs  
        Note (Just Fl) G -> Pitch.Gf 
        Note (Nothing) G -> Pitch.G 
        Note (Just Sh) G -> Pitch.Gs  
        Note (Just Fl) A -> Pitch.Af 
        Note (Nothing) A -> Pitch.A 
        Note (Just Sh) A -> Pitch.As  
        Note (Just Fl) B -> Pitch.Bf 
        Note (Nothing) B -> Pitch.B 
        Note (Just Sh) B -> Pitch.Bs  

classToRoot :: Pitch.Class -> Root
classToRoot clas = case clas of
        Pitch.Cf -> Note (Just Fl) C 
        Pitch.C  -> Note (Nothing) C          
        Pitch.Cs -> Note (Just Sh) C 
        Pitch.Df -> Note (Just Fl) D 
        Pitch.D  -> Note (Nothing) D          
        Pitch.Ds -> Note (Just Sh) D 
        Pitch.Ef -> Note (Just Fl) E 
        Pitch.E  -> Note (Nothing) E          
        Pitch.Es -> Note (Just Sh) E 
        Pitch.Ff -> Note (Just Fl) F 
        Pitch.F  -> Note (Nothing) F          
        Pitch.Fs -> Note (Just Sh) F 
        Pitch.Gf -> Note (Just Fl) G 
        Pitch.G  -> Note (Nothing) G          
        Pitch.Gs -> Note (Just Sh) G 
        Pitch.Af -> Note (Just Fl) A 
        Pitch.A  -> Note (Nothing) A          
        Pitch.As -> Note (Just Sh) A 
        Pitch.Bf -> Note (Just Fl) B 
        Pitch.B  -> Note (Nothing) B          
        Pitch.Bs -> Note (Just Sh) B 

degreeToAbs :: DiatonicDegree -> Int
degreeToAbs degree = case degree of 
        I   -> 0
        II  -> 1
        III -> 2
        IV  -> 3
        V   -> 4
        VI  -> 5
        VII -> 6
        o   -> 8


-- OLD: 
{-
rankSong :: [[(a,Double)]] -> [[((a,Double),Double)]]
rankSong (x:xs) = rankBar x:rankSong xs

rankBar :: [(a,Double)] -> [((a,Double),Double)]
rankBar bar = map boolToDouble bar

boolToDouble :: Bool -> Double
boolToDouble True  = 1.0
boolToDouble False = 0.0

fromScaleDegreeToRoot :: Chord (Note DiatonicDegree) -> Key -> Chord (Note DiatonicNatural)
fromScaleDegreeToRoot (Chord (Note (modifier) degree) shorthand add a b) (Key (Note (modifier2) n) mode) =
        (Chord (Note (modifier) root) shorthand add a b)
        where (Chord (Note (_) root)  _ _ _ _) = (keyChords (Key (Note (modifier2) n) mode))!!((diaDegToSemi(Note (modifier) degree))-1)
        
        -- this note map uses a circle of fifths based ordering
noteFifthsMaj :: [Note DiatonicNatural]
noteFifthsMaj =  [ Note Nothing   C -- "C"  0
                 , Note Nothing   F -- "F"  5
                 , Note (Just Fl) B -- "Bb" 10
                 , Note (Just Fl) E -- "Eb" 3
                 , Note (Just Fl) A -- "Ab" 8
                 , Note (Just Fl) D -- "Db" 1
                 , Note (Just Fl) G -- "Gb" 3
                 , Note (Just Fl) C -- "Gb" 3
                 , Note (Just Fl) F -- "Gb" 3
                 ]
-- diffent note names
noteFifthsMaj' :: [Note DiatonicNatural]
noteFifthsMaj' =  [ Note Nothing  C -- "B"  11
                 , Note Nothing   G -- "E"  4
                 , Note Nothing   D -- "A"  9
                 , Note Nothing   A -- "D"  2
                 , Note Nothing   E -- "G"  7
                 , Note Nothing   B -- "C"  0
                 , Note (Just Sh) F -- "Bb" 10
                 , Note (Just Sh) C -- "Eb" 3
                 , Note (Just Sh) G -- "Ab" 8
                 ] 
                  
-- this note map uses a circle of fifths based ordering
noteFifthsMin :: [Note DiatonicNatural]
noteFifthsMin =  [ Note Nothing   A -- "B"  11
                 , Note Nothing   E -- "E"  4
                 , Note Nothing   B -- "A"  9
                 , Note (Just Sh) F -- "D"  2
                 , Note (Just Sh) C -- "G"  7
                 , Note (Just Sh) G -- "C"  0
                 , Note (Just Sh) D -- "F"  5
                 , Note (Just Fl) B -- "Bb" 10
                 , Note Nothing   F -- "Eb" 3
                 , Note Nothing   C -- "Ab" 8
                 , Note Nothing   G -- "Eb" 3
                 , Note Nothing   D -- "Ab" 8
                 ]                 
-- different note names
noteFifthsMin' :: [Note DiatonicNatural]
noteFifthsMin' =  [ Note Nothing   A -- "B"  11
                 , Note Nothing   E -- "E"  4
                 , Note Nothing   B -- "A"  9
                 , Note (Just Fl) G -- "D"  2
                 , Note (Just Fl) D -- "G"  7
                 , Note (Just Fl) A -- "C"  0
                 , Note (Just Fl) E -- "F"  5
                 , Note (Just Sh) A -- "Bb" 10
                 , Note Nothing   F -- "Eb" 3
                 , Note Nothing   C -- "Ab" 8
                 , Note Nothing   G -- "Eb" 3
                 , Note Nothing   D -- "Ab" 8
                 ]  
                            
        
-}