-- Harmonizer 

{-# LANGUAGE CPP                      #-}
{-# LANGUAGE TemplateHaskell          #-}
{-# LANGUAGE TypeOperators            #-}
{-# LANGUAGE EmptyDataDecls           #-}
{-# LANGUAGE TypeSynonymInstances     #-}
{-# LANGUAGE FlexibleInstances        #-}
{-# LANGUAGE ScopedTypeVariables      #-}
{-# LANGUAGE TypeFamilies             #-}
{-# LANGUAGE GADTs                    #-}

module HarmTrace.Harmony.Main where
  
-- Music
import           HarmTrace.Base.MusicRep
import           HarmTrace.Models.Models
import           HarmTrace.Tokenizer.Tokenizer
import           HarmTrace.Harmony.ChordRep
import           HarmTrace.Harmony.HaskoreBridge
import           HarmTrace.HAnTree.HAnParser
import           HarmTrace.HarmTrace
import           HarmTrace.Audio.BeatChroma
-- UU Parsinglib
import           Text.ParserCombinators.UU
import           Text.ParserCombinators.UU.Demo.Examples
import           Text.ParserCombinators.UU.Utils
import           Text.ParserCombinators.UU.BasicInstances       hiding (Parser)
-- Data
import           Data.List
import           Data.Array
import           Data.Maybe
import           System.IO                
import qualified Numeric.NonNegative.Wrapper                    as NonNeg
import           Data.Ratio         
import           System.Random     
import           Main                                           hiding (main)      
import           HarmTrace.IO.Main
import           HarmTrace.HAnTree.ToHAnTree 
import           HarmTrace.IO.Errors 
import           HarmTrace.HAnTree.PostProcess
import Data.Ord (comparing)         
import          HarmTrace.HAnTree.Tree    
import          HarmTrace.HAnTree.HAn

type Harmony    = [ChordLabel]
type Octave     = Int   

modKeys :: [Key]
modKeys = [Key (Note (Just m) n ) a | m <- [Sh .. FF], n <-[C .. B], a <- [MajMode .. MinMode]]
normKeys :: [Key]
normKeys = [Key (Note (Nothing) n ) a | n <-[C .. B], a <- [MajMode .. MinMode]]
allKeys :: [Key]
allKeys = sort(modKeys ++ normKeys)

-- Usage: "harmonize [number of harmonizations as integer number] [key of the piece in HarmTrace notation] [path of input midi file as string]"
-- harmonize :: Int -> Key -> FilePath -> IO ()
-- harmonize amountOfHarms k path = do maybeconvert <- midi2HarmTrace path
main :: Int -> Key -> FilePath -> IO ()
main amountOfHarms k path = do maybeconvert <- midi2HarmTrace path
                               if (maybeconvert == Nothing) then do
                                        error ("harmonize: Error while opening file: " ++ path)
                                        return()
                                        else do let song = fromJust maybeconvert
                                                let (Song key t s m) = song    
                                                -- convert melody to relative ints
                                                let relativeMelody = barNotesToInts k m   
                                                -- make the chordlist with every bar
                                                let withChords = findCandidateChords k relativeMelody
                                                -- only chords on strong bar parts
                                                let skipped = filterSkipNotes t s withChords 
                                                -- give chords their propabilities
                                                let propped = propChords k skipped
                                                
                                                putStrLn ("--> Using file: " ++ path)                                    
                                                putStrLn ("--> Key: " ++ (keyToString k)) 
                                                putStrLn ("--> Time signature: " ++ (show t))
                                                putStrLn ("--> Melody:") 
                                                pretttyPrintMel propped         
                                                
                                                -- generate n random sequences  
                                                putStrLn $ "--> Creating " ++ (show amountOfHarms) ++ " harmonizations... "
                                                sequences <- generateNRandomSequences amountOfHarms propped
                                                let noImps =  map filterImpChords sequences
                                                let asRoots = (map.map) (chordScaleDegreeToRoot k) noImps                                                                        
                                                let grammar = Jazz
                                                let options = defaultOpts                                                                       
                                                
                                                -- as strings:
                                                let stringSeqs = map ((++) (keyToString k ++ ";")) $ init $ map sequenceAsString asRoots
                                                
                                                -- parse the strings all into a  list [ParseResult Piece]
                                                putStrLn $ "--> Parsing " ++ (show amountOfHarms) ++ " harmonizations... "
                                                parsedHarmonies <- mapM (parseTreeSilent grammar options) stringSeqs
                                                
                                                -- get the best of the parsed harmonies  
                                                bestErrorTrees <- getNBestParseResult 5 parsedHarmonies 
                                                
                                                putStrLn "--> 5 Best harmonizations: "
                                                -- show n best harmonisations
                                                showAllParseResults 0 bestErrorTrees
                                                
                                                let b@(ParseResult k chrd list tree ambtrees tokerr pieceerr pproc) = head bestErrorTrees
                                                --printParseresult b         
                                                putStrLn ("--> Chords: " ++ (show chrd))
                                                
                                                -- make a midi file from the melody and harmonization
                                                putStrLn ("--> Writing midi file: " ++ (filenameAddHarm path))
                                                toMidi t s (filenameAddHarm path) m chrd
                                                                                               
                                                return()                                              

-- Usage: "harmonize [number of harmonizations as integer number] [key of the piece in HarmTrace notation] [path of input midi file as string]"
--harmonize :: FilePath -> IO()
harmonize2 amountOfHarms k l path = do maybeconvert <- midi2HarmTrace path
                                       if (maybeconvert == Nothing) then do
                                              error ("harmonize: Error while opening file: " ++ path)
                                              return()
                                              else do let song = fromJust maybeconvert
                                                      let (Song key t l m) = song    
                                                      -- convert melody to relative ints
                                                      let relativeMelody = barNotesToInts k m   
                                                      putStrLn ("rel. mel: " ++ (show relativeMelody))
                                                      -- make the chordlist with every bar
                                                      let withChords = findCandidateChords k relativeMelody
                                                      putStrLn ("mel with chords " ++ (show withChords))
                                                      -- only chords on strong bar parts
                                                      --let skipped = filterSkipNotes t l withChords 
                                                      let skipped = withChords
                                                      putStrLn ("skipped " ++ (show skipped))
                                                      -- give chords their propabilities
                                                      let propped = propChords k skipped
                                                      
                                                      putStrLn ("--> Using file: " ++ path)                                    
                                                      putStrLn ("--> Key: " ++ (keyToString k)) 
                                                      putStrLn ("--> Time signature: " ++ (show t))
                                                      putStrLn ("--> Melody:") 
                                                      pretttyPrintMel propped         
                                                      
                                                      -- generate n random sequences  
                                                      putStrLn $ "--> Creating " ++ (show amountOfHarms) ++ " harmonizations... "
                                                      sequences <- generateNRandomSequences amountOfHarms propped
                                                      let noImps =  map filterImpChords sequences
                                                      let asRoots = (map.map) (chordScaleDegreeToRoot k) noImps                                                                        
                                                      let grammar = Jazz
                                                      let options = defaultOpts                                                                       
                                                      
                                                      -- as strings:
                                                      let stringSeqs = map ((++) (keyToString k ++ ";")) $ init $ map sequenceAsString asRoots
                                                      
                                                      -- parse the strings all into a  list [ParseResult Piece]
                                                      putStrLn $ "--> Parsing " ++ (show amountOfHarms) ++ " harmonizations... "
                                                      parsedHarmonies <- mapM (parseTreeSilent grammar options) stringSeqs
                                                      
                                                      -- get the best of the parsed harmonies  
                                                      bestErrorTrees <- getNBestParseResult 5 parsedHarmonies 
                                                      
                                                      putStrLn "--> 5 Best harmonizations: "
                                                      -- show n best harmonisations
                                                      showAllParseResults 0 bestErrorTrees
                                                      
                                                      let b@(ParseResult k chrd list tree ambtrees tokerr pieceerr pproc) = head bestErrorTrees
                                                      --printParseresult b         
                                                      putStrLn ("--> Chords: " ++ (show chrd))
                                                      
                                                      -- make a midi file from the melody and harmonization
                                                      putStrLn ("--> Writing midi file: " ++ (filenameAddHarm path))
                                                      toMidi t l (filenameAddHarm path) m chrd
                                                                                                     
                                                      return() 

-- adds "_harmonized" and the current time to the filename
filenameAddHarm :: FilePath -> FilePath
filenameAddHarm path | ((drop (length(path)-4) path) == ".mid")  = ((take (length(path)-4) path) ++ "_harmonized.mid")
                     | ((drop (length(path)-5) path) == ".midi") = ((take (length(path)-5) path) ++ "_harmonized.midi")
                     | otherwise                                  = (path ++ "_harmonized.midi")
                     
printParseresult :: ParseResult g -> IO ()
printParseresult p@(ParseResult k chrd list tree ambtrees tokerr pieceerr pproc) = 
        do putStrLn ("Key : " ++ (show k))
           putStrLn ("Chords : " ++ (show chrd))
           putStrLn ("countErrors : " ++ (show $ countErrors pieceerr))
           putStrLn ("Tree : " ++ (show chrd))           
           putStrLn ("Func Nodes : " ++ (show $ getNoFuncNodes p))           
           putStrLn ("Ambtrees : " ++ (show ambtrees))
           putStrLn ("Token err : " ++ (show $ length tokerr))
           putStrLn ("Parser err : " ++ (show $ length pieceerr))  
           putStrLn ("postProcessing : " ++ (show postProcessing))
           return()
           
getByLength :: [ParseResult g] -> [ParseResult g]
getByLength [] = []
getByLength (p@(ParseResult k chrd list tree ambtrees tokerr pieceerr pproc):xs) | ((length chrd) == getNrFuncNodes tree) = p:getByLength xs
                                                                                 | otherwise                              = getByLength xs

-- returns the n best parse results by 
--   first sorting on the least amount of general errors
getNBestParseResult :: (GTree g) => Int -> [ParseResult g] -> IO [ParseResult g]
getNBestParseResult n a = do putStrLn ("--> Sorting harmonizations...")
                             let bestErrorTrees = take n $ sortBy sortParseResult3 a
                             let treesByNoNodes = sortBy (comparing getNoFuncNodes) bestErrorTrees
                             return(treesByNoNodes)          

getNoFuncNodes :: (ParseResult a) -> Int
getNoFuncNodes (ParseResult _ _ _ t _ _ _ _) = getNoFuncNode t where
        getNoFuncNode :: Tree HAn -> Int
        getNoFuncNode (Node (HAnFunc P) nodes _) = length nodes
        getNoFuncNode _ = error "HarmTrace.hs: not a correctly formed HAn Tree"                              

getBestParseResult :: [ParseResult g] -> ParseResult g
getBestParseResult a = head $ sortBy sortParseResult a                                                        

-- sort by deletions
sortParseResult (ParseResult k chrd list tree ambtrees tokerr pieceerr pproc) (ParseResult k2 chrd2 list2 tree2 ambtrees2 tokerr2 pieceerr2 pproc2) 
        | d < d2 = GT   
        | d >= d2 = LT  where 
                (ErrorNrs i d de r) = countErrors pieceerr
                (ErrorNrs i2 d2 de2 r2) = countErrors pieceerr2

-- sort by insertions
sortParseResult2 (ParseResult k chrd list tree ambtrees tokerr pieceerr pproc) (ParseResult k2 chrd2 list2 tree2 ambtrees2 tokerr2 pieceerr2 pproc2) 
        | (length chrd) >  (length chrd2) = GT   
        | (length chrd) <= (length chrd2) = LT  where 
                (ErrorNrs i d de r) = countErrors pieceerr
                (ErrorNrs i2 d2 de2 r2) = countErrors pieceerr2   

-- sort by simple error simpleErrorMeasure
sortParseResult3 (ParseResult k chrd list tree ambtrees tokerr pieceerr pproc) (ParseResult k2 chrd2 list2 tree2 ambtrees2 tokerr2 pieceerr2 pproc2) 
        | (simpleErrorMeasure e1) >  (simpleErrorMeasure e2) = GT   
        | (simpleErrorMeasure e1) <= (simpleErrorMeasure e2) = LT  where 
                e1 = countErrors pieceerr
                e2 = countErrors pieceerr2        
                
showAllParseResults :: (GTree g) => Int -> [ParseResult g] -> IO()
showAllParseResults _ []     = return()
showAllParseResults n (x:xs) = do putStr $ show n
                                  mapM_ (print . gTreeHead) (parsedPiece x) 
                                  printParseresult x
                                  showAllParseResults (n+1) xs
                                  return()

sequenceAsString :: [Chord Root] -> String          
sequenceAsString []                                 = ""
sequenceAsString [(Chord (Note (Nothing) r) sh l i d)]    = ((show r) ++ ":" ++ (shortHandAsString sh))
sequenceAsString ((Chord (Note (Nothing) r) sh l i d):xs) = ((show r) ++ ":" ++ (shortHandAsString sh)) ++ ";" ++ sequenceAsString xs
sequenceAsString [(Chord (Note (m) r) sh l i d)]    = ((show r) ++ show(fromJust m) ++ ":" ++ (shortHandAsString sh))
sequenceAsString ((Chord (Note (m) r) sh l i d):xs) = ((show r) ++ show(fromJust m) ++ ":" ++ (shortHandAsString sh)) ++ ";" ++ sequenceAsString xs


shortHandAsString :: Shorthand -> String
shortHandAsString  Maj = "maj"
shortHandAsString  Min = "min"
shortHandAsString  Dim = "dim"
shortHandAsString  Aug = "aug"
shortHandAsString  Maj7 = "maj7"
shortHandAsString  Min7 = "min7"
shortHandAsString  Sev = "7"
shortHandAsString  Dim7 = "dim7"
shortHandAsString  HDim7 = "hdim"
shortHandAsString  MinMaj7 = "minmaj7"
shortHandAsString  Maj6 = "maj6"
shortHandAsString  Min6 = "min6"
shortHandAsString  Nin  = "9"
shortHandAsString  Maj9 = "maj9"
shortHandAsString  Min9 = "min9"
shortHandAsString  Sus4 = "sus4" 

filterImpChords :: [Chord ScaleDegree] -> [Chord ScaleDegree]
filterImpChords []     = []
filterImpChords (x:xs) | (x == (Chord (Note (Nothing) Imp) None [] (-9) (-9))) = filterImpChords xs
                       | otherwise                                             = x:filterImpChords xs
                   
generateNRandomSequences :: Int -> [(Bar,[(Int,Octave,Duration,PropChordList)])] -> IO [[Chord ScaleDegree]]
generateNRandomSequences 0 l  = return([[]])
generateNRandomSequences _ [] = return([[]])
generateNRandomSequences n l  = do chordSequence <- generateRandomSequence l
                                   chordList <- generateNRandomSequences (n-1) l
                                   return(chordSequence:chordList)

generateRandomSequence :: [(Bar,[(Int,Octave,Duration,PropChordList)])] -> IO [Chord ScaleDegree]
generateRandomSequence [] = return([])
generateRandomSequence ((b,m):ms) = do chords <- pickChord m 
                                       next <- generateRandomSequence ms
                                       return(chords ++ next)

pickChord :: [(Int,Octave,Duration,PropChordList)] -> IO [Chord ScaleDegree]
pickChord []             = return([])
pickChord ((i,o,d,l):ms) = do a <- (pick (sortPropChordList (transf l)))
                              b <- pickChord ms
                              return(a:b)

pick :: PropChordList -> IO (Chord ScaleDegree)
pick []                        = return(Chord (Note (Nothing) Imp) None [] (-9) (-9))
pick [(c1,p1),(c2,p2),(c3,p3)] = do prob <- getStdRandom (randomR (0.0,1.0))
                                    if (prob <= p3) 
                                            then do return(c3)
                                            else if ((prob > p3) && (prob <= (p3+p2))) 
                                                    then do return(c2)
                                                    else if ((prob > (p3+p2)) && (prob <= 1))
                                                            then do return(c1)
                                                            else return(c1)
pick [(c1,p1),(c2,p2)]         = do prob <- getStdRandom (randomR (0.0,1.0))
                                    if (prob >= p1) 
                                            then do return(c1)
                                            else return(c2)

pick [(c1,p1)]                 = do return(c1)    

sortPropChordList :: PropChordList -> PropChordList
sortPropChordList l = sortBy sortGT l    

sortGT (a1, b1) (a2, b2)
  | b1 < b2 = GT
  | b1 >= b2 = LT

transf :: PropChordList -> PropChordList
transf []         = []
transf [(c1,p1),(c2,p2),(c3,p3)] = [(c1,np1),(c2,np2),(c3,np3)] where 
                                           np1 = (p1/(p1+p2+p3))
                                           np2 = (p2/(p1+p2+p3))
                                           np3 = (p3/(p1+p2+p3))                                                                                      
transf [(c1,p1),(c2,p2)] = [(c1,np1),(c2,np2)] where
                                           np1 = (p1/(p1+p2))
                                           np2 = (p2/(p1+p2))
transf [(c,_)] = [(c,1)]

propChords :: Key -> [(Bar,[(Int,Octave,Duration,ChordList)])] -> [(Bar,[(Int,Octave,Duration,PropChordList)])]
propChords k []         = []
propChords k ((b,m):ms) = (b,(propMel k m)):propChords k ms where
        propMel :: Key -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,PropChordList)] 
        propMel k []             = []
        propMel k ((i,o,d,l):ms) = (i,o,d,(prop k l i)):propMel k ms where
                prop :: Key -> ChordList -> Int -> PropChordList
                prop _ []     _ = []
                prop k (x:xs) i = (x,(getProp k x i)): prop k xs i

pretttyPrintMel :: (Show a) => (Show b) => [(Bar,[(a,Octave,Duration,b)])] -> IO ()
pretttyPrintMel ((b,l):xs) = do putStrLn ("    " ++ (show(b)) ++ " " ++ show (l) )
                                pretttyPrintMel xs
pretttyPrintMel []         = return ()     

pretttyPrintSeq :: (Show a) => [[a]] -> IO ()
pretttyPrintSeq (x:xs) = do putStrLn (show(x))
                            pretttyPrintSeq xs
pretttyPrintSeq []         = return ()     

barNotesToInts :: Key -> [(Bar,[MelodyTone])] -> [(Bar,[(Int,Octave,Duration,ChordList)])]
barNotesToInts k ((b,m):xs) = (b,(melodyToInt k m)): barNotesToInts k xs where
barNotesToInts _ []           = []      

melodyToInt :: Key -> [(Root,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
melodyToInt key ((n,o,d,l):notes) = ((diaDegToSemi $ (toScaleDegree key n)),o,d,l):melodyToInt key notes
melodyToInt key []              = [] 

findCandidateChords :: Key -> [(Bar,[(Int,Octave,Duration,ChordList)])] -> [(Bar,[(Int,Octave,Duration,ChordList)])]
findCandidateChords k []               = []  
findCandidateChords k ((b@(1,_),m):ms) = (b,(f k m)) : findCandidateChords k ms where
                                         f :: Key -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
                                         f k ((i,o,d,l):xs) = (i,o,d,(onlyVOrI $ findChordsByInt2 i k 0)): f' k xs
                                         f _ []           = []
                                         f' :: Key -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
                                         f' k ((i,o,d,l):xs) = (i,o,d,(findChordsByInt2 i k 0)): f' k xs
                                         f' _ []           = []                                               
-- last chord of last bar = I or V
-- if n-1 can be V when bar n = I, it must be V                                         
findCandidateChords k ((b@(n,d),m):ms) | (n==d)    =  (b,(f2 k m)) : findCandidateChords k ms 
                                       | otherwise =  (b,(f' k m)) : findCandidateChords k ms where
                                                        f' :: Key -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
                                                        f' k ((i,o,d,l):xs) = (i,o,d,(findChordsByInt2 i k 0)): f' k xs
                                                        f' _ []           = []   
                                                        f2 :: Key -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
                                                        f2 k ((i,o,d,l):[]) = [(i,o,d,(onlyVOrI $ (findChordsByInt2 i k 0)))]         
                                                        f2 k ((i,o,d,l):[(i2,o2,d2,l2)]) = [(i,o,d,(checkIV l2 $ (findChordsByInt2 i k 0)))] -- !
                                                        f2 k ((i,o,d,l):xs) = (i,o,d,(findChordsByInt2 i k 0)): f2 k xs
                                                        f2 _ []           = []

chordElem :: DiatonicDegree -> ChordList -> Bool               
chordElem _ []                                 = False
chordElem d ((Chord (Note (_) sd) _ _ _ _):xs) | (sd == d) = True
                                               | otherwise = chordElem d xs                                               
                                               
checkIV :: ChordList -> ChordList -> ChordList
checkIV [] []                                = []  
checkIV this@(x:xs) next@(y:ys)              | (chordElem V next) = onlyIV this
                                             | otherwise          = this

onlyIV :: ChordList -> ChordList
onlyIV []                                   = []
onlyIV (c@(Chord (Note (_) sd) _ _ _ _):xs) | (sd == IV)  = c:onlyVOrI xs
                                            | otherwise   = onlyVOrI xs  

onlyVOrI :: ChordList -> ChordList
onlyVOrI []                                   = []
onlyVOrI (c@(Chord (Note (_) sd) _ _ _ _):xs) | ((sd == I) || (sd == V))  = c:onlyVOrI xs
                                              | otherwise                 = onlyVOrI xs 
onlyI :: ChordList -> ChordList
onlyI []                                   = []
onlyI (c@(Chord (Note (_) sd) _ _ _ _):xs) | ((sd == I))               = c:onlyI xs
                                           | otherwise                 = onlyI xs   

-- give every chord the number it occurs in the list and remove doubles
occursCheck :: (Eq a) => [a] -> [(a, Double)]
occursCheck l = occursCheck' l l where 
        occursCheck' :: (Eq a) => [a] -> [a] -> [(a, Double)]
        occursCheck' (x:xs) l2 = (x,fromIntegral(length(intersect l2 [x]))):occursCheck' xs l2
        occursCheck' []     _  = []

filterSkipNotes :: TimeSig -> Int -> [(Bar,[(Int,Octave,Duration,ChordList)])] -> [(Bar,[(Int,Octave,Duration,ChordList)])] 
-- filterSkipNotes t 1 l              = l
filterSkipNotes t s (((n,d),m):ms) = ((n,d),(filterSkipBar t s 0 m)):filterSkipNotes t s ms
filterSkipNotes t s []             = [] 
 
filterSkipBar :: TimeSig -> Int -> Rational -> [(Int,Octave,Duration,ChordList)] -> [(Int,Octave,Duration,ChordList)]
filterSkipBar (b1,b2) s a ((i,o,d,l):ms) | (a == (toRational(s) / toRational(b2)) || (a == 0)) = (i,o,d,l ):filterSkipBar (b1,b2) s (a+d) ms
                                         | otherwise                                           = (i,o,d,[]):filterSkipBar (b1,b2) s (a+d) ms
filterSkipBar s _ _       []             = []

keyToString :: Key -> String
keyToString  (Key r m) = show r ++ modeToString m

modeToString :: Mode -> String
modeToString m | (null $ show m) = ":maj"
               | otherwise       = ":min"
                                        
{- 

-- OLD functions that uses the text input for a melody
harmonizeSong :: [[(Int,Double)]] -> Key -> Int -> [[[Chord ScaleDegree]]] 
harmonizeSong [] key          _ =  []
harmonizeSong (bar1:bars) key i =  (harmonizeBar bar1 key 0):harmonizeSong bars key (i+1) 
              where 
              harmonizeBar :: [(Int,Double)] -> Key -> Int -> [[Chord ScaleDegree]]
              harmonizeBar [] key           i = []
              harmonizeBar (note:notes) key i = (findChordsByInt note key i):harmonizeBar notes key (i+1)

melodyToScaleDegree :: Key -> [(Root,Double)] -> [(ScaleDegree,Double)]
melodyToScaleDegree key ((note,dur):notes) = (toScaleDegree key (note),dur) : melodyToScaleDegree key (notes)
melodyToScaleDegree key []                 = []

scaledegreeToSemi :: [(ScaleDegree,Double)] -> [(Int,Double)]
scaledegreeToSemi ((note,dur):notes) = (diaDegToSemi (note),dur) : scaledegreeToSemi (notes)
scaledegreeToSemi []                 = []

phrasingMelody :: Int -> String -> String
phrasingMelody beat melody = setToBeat beat melody 0

seperateMelodyByBeat :: [(a,Double)] -> Beat ->  [[(a,Double)]]
seperateMelodyByBeat melody beat = map reverse $ ph2 melody beat [] []

setToBeat :: Int -> String -> Int -> String 
setToBeat beat (x:xs) a | ((a `mod` beat) == 0) =  x  : setToBeat beat xs (a+1)
                        | otherwise             = '-' : setToBeat beat xs (a+1)
setToBeat beat [] a     = []
           
harmonizerText :: FilePath -> IO ()
harmonizerText filenam2 = do    file <- openFile filename ReadMode
                                contents <- hGetContents file
                                (Song beat key melody contents) <- makeSong contents                         
                                putStrLn ("--> Using file: " ++ filename)
                                putStrLn ("--> Beat is: " ++ show(beat))
                                putStrLn ("--> Key is : " ++ show(key))
                                putStrLn ("--> Melody after parsing is : " ++ show(melody))                         
                                let relative = melodyToScaleDegree key melody
                                --putStrLn ("--> Relative tones from melody : " ++ show(relative))
                                let relativeints = scaledegreeToSemi relative
                                putStrLn ("--> As Ints : " ++ show(relativeints))                         
                                -- Phrased melody
                                let phrased = seperateMelodyByBeat relativeints beat
                                --putStrLn ("--> After setting the notes to the beat the melody is : " ++ show(phrased))
                                -- Possible Harmonizations:
                                let harmonized = harmonizeSong phrased key 0
                                --putStrLn ("--> Possible harmonisations are: " ++ show(harmonized))                         
                                let chanced = songChordProps key phrased harmonized
                                putStrLn ("--> Chances for the chords: " ++ show(chanced))
                                let bestharm = beatFilter beat (getBestHarm chanced) 0
                                putStrLn ("--> Best harmonization: " ++ show(bestharm))                         
                                let rootchords = map (chordScaleDegreeToRoot key) bestharm
                                putStrLn ("--> Best harmonization as roots: " ++ show(rootchords))                         
                                let noImpChords = filter (not.isNoneChord) rootchords
                                putStrLn ("--> No imps: " ++ show(noImpChords))                                         
                                --let hkchords = map htChord2hkChord noImpChords
                                putStrLn ("--> Making MIDI file: " ++ filenam2)
                                htToMidi filenam2 melody noImpChords                         
                                return()  
 
-}
-- OLD: 

{-

stringToNotes :: [String] -> [Note DiatonicNatural]
stringToNotes (x:xs) = stringToNote x : stringToNotes xs
stringToNotes []     = []

noteToInt :: Note DiatonicNatural ->  [Note DiatonicNatural] -> Int -> String
noteToInt note (x:xs) a | (note == x) = show(a)
                        | otherwise   =  noteToInt note xs (a+1)
noteToInt note []     a = show(9) -- 9 means error

barNotesToScaleDegree :: Key -> [(Beat,ChordList,[MelodyTone])] -> [(Beat,ChordList,[(ScaleDegree,Octave,Duration)])]
barNotesToScaleDegree k ((b,c,m):xs) = (b,c,(melodyToScaleDegree k m)): barNotesToScaleDegree k xs where
barNotesToScaleDegree _ []           = [] 

melodyToScaleDegree :: Key -> [(Root,Octave,Duration)] -> [(ScaleDegree,Octave,Duration)]
melodyToScaleDegree key ((n,o,d):notes) = ((toScaleDegree key n),o,d):melodyToScaleDegree key notes
melodyToScaleDegree key []              = []    

-- let s ="B:maj;B:maj;B:maj;E:maj;F#:maj;G#:min;G#:min;E:maj;C#:min;F#:maj;F#:maj;B:maj"
-- tr <- parseTree Pop defaultOpts s
-- mapM_ (print . gTreeHead) (parsedPiece tr) 
-- [Piece[PT[T_1_1[I_0[B:Maj][B:Maj]]]][PD[D_1_1[S_3_1[IV_0[E:Maj]]][D_3_1[V_0[F#:Maj]]]]][PD[D_1_1[S_5par_1[VIm_0[G#:Min][G#:Min]]][D_1_1[S_3_1[IV_0[E:Maj]]][D_1_1[S_1par_1[IIm_0[C#:Min]]][D_3_1[V_0[F#:Maj][F#:Maj]]]]]]][PT[T_1_1[I_0[B:Maj]]]]]


-}
