{- This module contains the parsers needed to parse an 
   inputfile with a key, time signature and melody into 
   a Song datatype 
   This is no longer used -}
   
module HarmTrace.Harmony.Parser where

import HarmTrace.Base.MusicRep
import HarmTrace.Tokenizer.Tokenizer
-- Parser
import Text.ParserCombinators.UU
import Text.ParserCombinators.UU.Demo.Examples
import Text.ParserCombinators.UU.Utils
import Text.ParserCombinators.UU.BasicInstances hiding (Parser)

data Song a = Song { beat         :: Beat
                   , key          :: Key
                   , melody       :: [(a,Double)] -- was: :: [(Root,Double)]     
                   , filestring   :: String       -- was: filestring   :: a     
                   }     

type Harmony = [ChordLabel]
type Beat = (Double,Double)

makeSong :: String -> IO (Song Root)
makeSong input = do let ((b,b2),key,melody) = doParse pSong input
                    return (Song (read b :: Double,read b2 :: Double) key melody input)

pSong :: Parser ((String,String),Key,[(Root,Double)])
pSong = (\b k m -> (b,k,m)) <$> pBeat <*> parseKey <*> pMelody

pHarmonyMaj :: Parser [Chord ScaleDegree]
pHarmonyMaj =  pMany (note2ChordMaj <$> (lift <$> pDigit)) <* pEnd

pHarmonyMin :: Parser [Chord ScaleDegree]
pHarmonyMin =  pMany (note2ChordMin <$> (lift <$> pDigit)) <* pEnd

pFile :: Parser [Note DiatonicNatural]  
-- was : pMany (stringToNote <$> pNote) <* pEnd
pFile = pMany (string2NoteDia <$> pNote) <* pEnd

pMelody :: Parser [(Root,Double)]
pMelody = pMany ( (pPacked (pSym '(') (pSym ')')) ((\a b -> (a,b)) <$> parseRootComma <*> pDouble) ) <* pEnd

parseRootComma :: Parser Root
parseRootComma = parseRoot <* pComma
                          
pNote :: Parser String 
pNote = pToken "C " <|> pToken "C# " <|> pToken "Cb " <|> pToken "C## " <|> pToken "Cbb " <|>
        pToken "D " <|> pToken "D# " <|> pToken "Db " <|> pToken "D## " <|> pToken "Dbb " <|> 
        pToken "E " <|> pToken "E# " <|> pToken "Eb " <|> pToken "E## " <|> pToken "Ebb " <|>
        pToken "F " <|> pToken "F# " <|> pToken "Fb " <|> pToken "F## " <|> pToken "Fbb " <|>
        pToken "G " <|> pToken "G# " <|> pToken "Gb " <|> pToken "G## " <|> pToken "Gbb " <|>
        pToken "A " <|> pToken "A# " <|> pToken "Ab " <|> pToken "A## " <|> pToken "Abb " <|>
        pToken "B " <|> pToken "B# " <|> pToken "Bb " <|> pToken "B## " <|> pToken "Bbb "
      
pBeat :: Parser (String,String)
pBeat = (\x y -> (x,y)) <$> (lift <$> pDigit) <*> pBeat'  <|>
         (\x y -> (x,y)) <$> ((\a b ->(a++b)) <$> (lift <$> pDigit) <*> (lift <$> pDigit)) <*> pBeat'

pBeat' :: Parser String
pBeat' = (pToken "/") *> (lift <$> pDigit)

pKeyFromFile :: Parser Key
pKeyFromFile = (pBeat *> pNl *> parseKey <* pNl)

pNl :: Parser String
pNl = pToken "\n"

doParse :: Parser t -> String -> t
doParse p inp = parse (p) (createStr (LineColPos 0 0 0) inp) 

note2ChordMaj :: String -> Chord ScaleDegree
note2ChordMaj m = case m of 
                       "0" -> Chord (Note (Nothing) I) Maj [] 0 0
                       "1" -> Chord (Note (Nothing) II) Min [] 0 0
                       "2" -> Chord (Note (Nothing) III) Min [] 0 0
                       "3" -> Chord (Note (Nothing) IV) Maj [] 0 0
                       "4" -> Chord (Note (Nothing) V) Maj [] 0 0
                       "5" -> Chord (Note (Nothing) III) Min [] 0 0
                       "6" -> Chord (Note (Nothing) VII) Maj [] 0 0
                       o  -> Chord (Note (Nothing) Imp) None [] 0 0
                                                
note2ChordMin :: String -> Chord ScaleDegree
note2ChordMin m  = case m of 
                        "0" -> Chord (Note (Nothing) I) Maj [] 0 0
                        "1" -> Chord (Note (Nothing) II) Min [] 0 0
                        "2" -> Chord (Note (Nothing) III) Min [] 0 0
                        "3" -> Chord (Note (Nothing) IV) Maj [] 0 0
                        "4" -> Chord (Note (Nothing) V) Maj [] 0 0
                        "5" -> Chord (Note (Nothing) III) Min [] 0 0
                        "6" -> Chord (Note (Nothing) VII) Maj [] 0 0
                        o  -> Chord (Note (Nothing) Imp) None [] 0 0

-- Converting a String to a Note DiatonicNatural
string2NoteDia :: String -> (Note DiatonicNatural)
string2NoteDia [r]   = (Note (Nothing) root) where 
                              root     = (read [r]    :: DiatonicNatural)
string2NoteDia s     = (Note (modifier) root) where 
                              modifier = Just (read (tail(s))    :: Modifier)
                              root     = (read (lift (head(s)))  :: DiatonicNatural)

-- Phrasing
ph2 :: [(a,Double)] -> Beat -> [Double] -> [(a,Double)] -> [[(a,Double)]]
ph2 ((note,dur):xs) (b,b2) a a2 | (sum(a) /= (b / b2)) = ph2 xs (b,b2) (dur:a) ((note,dur):a2)
                                | otherwise            = a2:(ph2 ((note,dur):xs) (b,b2) [] [])
ph2 [] n a a2                   = [a2]

ph :: [Int] -> Int -> [[Int]]
ph (x:xs) n = map reverse $ ph' (x:xs) n []

ph' :: [Int] -> Int -> [Int] -> [[Int]]
ph' (x:xs) n a | (sum(a) /= n) = ph' xs n (x:a)
               | otherwise     = a:(ph' (x:xs) n [])
ph' [] n a     = [a]


tupleMap :: (a -> c) -> [(a,b)] -> [(c,b)]
tupleMap f ((x,y):xs) = ((f x),y):tupleMap f xs
tupleMap f []         = []
        
--- OLD
{-
pBeat :: Parser String
pBeat = ((lift <$> pDigit) <* (pToken "/") <* (lift <$> pDigit)) <|>
        ((\a b ->(a++b)) <$> (lift <$> pDigit) <*> (lift <$> pDigit) <* (pToken "/") <* (lift <$> pDigit))
 
stringToNote :: String -> (Note DiatonicNatural)
stringToNote (x:xs) = if (modifier == No) then 
                         (Note (Nothing) root)
                      else  
                         (Note (Just modifier) root)
                      where root     = read [x] :: DiatonicNatural 
                            modifier = read (init xs) :: Modifier 
                            
-}